#!/usr/bin/env gosh

;; Copyright (c) Iwata <iwata@quasiquote.org>
;;
;; Permission is hereby granted, free of charge, to any person obtaining
;; a copy of this software and associated documentation files (the
;; "Software"), to deal in the Software without restriction, including
;; without limitation the rights to use, copy, modify, merge, publish,
;; distribute, sublicense, and/or sell copies of the Software, and to
;; permit persons to whom the Software is furnished to do so, subject to
;; the following conditions:
;;
;; The above copyright notice and this permission notice shall be included
;; in all copies or substantial portions of the Software.
;;
;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
;; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
;; MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
;; IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
;; CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
;; TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
;; SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

(define-module text.mijinko
  (use srfi-1)
  (use srfi-27)
  (use dbm)
  (use util.list)
  (use text.kakasi)
  (use mecab)
  (export mijinko-dbtype mijinko-gram mijinko-prompt
          mijinko-main))

(select-module text.mijinko)

(define mijinko-dbtype #f)

(define mijinko-gram 3)

(define (mijinko-your-prompt)
  (display "> ")(flush))

(define (mijinko-prompt)
  (display "mijinko> "))

(define mijinko-random-source #f)
(define (mijinko-random n)
  (unless mijinko-random-source
    (begin
      (set! mijinko-random-source (make-random-source))
      (random-source-randomize! mijinko-random-source)))
  ((random-source-make-integers mijinko-random-source) n))

(define mijinko-wakati-method 'mecab)
(define mijinko-mecab #f) ;;; XXX cannot put in (let ... SEGV
(define (mijinko-wakati)
  (case mijinko-wakati-method
    ((mecab)
     (lambda (s)
       (unless mijinko-mecab
         (set! mijinko-mecab (mecab-new2 "")))
       (drop-right ;; drop last EOS
        (string-split (mecab-sparse-tostr mijinko-mecab s)
                      #/\t[^\n]*\n/) 1)))
    ((kakasi) (kakasi-wakati))))

;;
;; db-operations
;;
;; db structure have key and item pair.
;; key and item are s-expression.
;;
;;  key: atom (1st word)
;;  item: alist ((another-words . count) ..)
;;

(define (mijinko-restore gram db)
  (let* ((key (car gram))
         ;; item: list
         (item (cdr gram))
         ;; db-item: alist
         (db-item (read-from-string
                   (dbm-get db (write-to-string key) ""))))
    (if (eof-object? db-item)
        ;; append new item
        (dbm-put! db (write-to-string key)
                  (write-to-string `((,item . 1))))
        ;; take item
        (let* ((val (assoc item db-item))
               (count (if val (cdr val) 0))
               ;; increment count
               (new-item (assoc-set! db-item item (+ 1 count))))
          ;; restore
          (dbm-put! db (write-to-string key)
                    (write-to-string new-item))))))

(define (mijinko-learn grams db)
  (define (loop l)
    (cond ((< (length l) mijinko-gram)
           ;; finish
           #t)
          (else
           (let ((gram (take l mijinko-gram)))
             (mijinko-restore gram db)
             (loop (cdr l))))))
  (loop (append '(*TOP*) grams '(*END*))))

;; return one list from statistical rules
(define (mijinko-select-gram item)
  (let* ((denom (apply + (map cdr item))) ;; take whole count
         (num (mijinko-random denom)))
    (define (loop l n)
      (if (<= num (+ n (cdar l)))
          (caar l)
          (loop (cdr l) (+ n (cdar l)))))
    (loop item 0)))

(define (mijinko-chat db)
  (define (loop l buf)
    (let ((latest (take-while (lambda (s) (not (eq? s '*END*))) l)))
      (if (find (lambda (s) (eq? s '*END*)) l)
          (string-append (string-join (reverse buf) "")
                         (string-join latest ""))
          (loop (mijinko-select-gram
                 (read-from-string (dbm-get db (write-to-string (car l)) "")))
                (cons (car l) buf)))))
  (loop
   (mijinko-select-gram
    (read-from-string (dbm-get db (write-to-string '*TOP*))))
   '()))

(define (mijinko-learn-and-chat str learn quiet verbose db)
  (let ((grams ((mijinko-wakati) str)))
    (if learn
        (mijinko-learn grams db))
    (unless quiet
      (begin
        (mijinko-prompt)
        (print (mijinko-chat db))))))

;; main loop
(define (mijinko-main learn quiet verbose db-name)
  (let* ((*db* (if learn
                   (dbm-open mijinko-dbtype :path db-name :rw-mode :write)
                   (dbm-open mijinko-dbtype :path db-name :rw-mode :read))))
    (define (loop)
      ;; prompt
      (unless quiet
        (mijinko-your-prompt))
      ;; take one line
      (let* ((s (read-line (standard-input-port))))
        (cond ((eof-object? s)
               #t)
              ;; ignore blank line
              ((#/^[\s\t]*$/ s)
             (loop))
              (else
               (mijinko-learn-and-chat s learn quiet verbose *db*)
               (loop)))))
    (loop)))

(provide "text/mijinko")
